/* Low level AX.25 code:
 *  incoming frame processing (including digipeating)
 *  IP encapsulation
 *  digipeater routing
 *
 * Copyright 1991 Phil Karn, KA9Q
 */
#include <stdio.h>
#include "global.h"
#include "mbuf.h"
#include "iface.h"
#include "arp.h"
#include "slip.h"
#include "ax25.h"
#include "lapb.h"
#include "netrom.h"
#include "ip.h"
#include "devparam.h"
#include <ctype.h>
#include "lapb.h"

static int axsend(struct iface *iface,char *dest,char *source,
	int cmdrsp,int ctl,struct mbuf *data);

/* List of AX.25 multicast addresses in network format (shifted ascii).
 * Only the first entry is used for transmission, but an incoming
 * packet with any one of these destination addresses is recognized
 * as a multicast.
 */
char Ax25multi[][AXALEN] = {
	'Q'<<1, 'S'<<1, 'T'<<1, ' '<<1, ' '<<1, ' '<<1, '0'<<1, /* QST    */
	'M'<<1, 'A'<<1, 'I'<<1, 'L'<<1, ' '<<1, ' '<<1, '0'<<1, /* MAIL   */
	'N'<<1, 'O'<<1, 'D'<<1, 'E'<<1, 'S'<<1, ' '<<1, '0'<<1, /* NODES  */
	'I'<<1, 'D'<<1, ' '<<1, ' '<<1, ' '<<1, ' '<<1, '0'<<1, /* ID     */
	'O'<<1, 'P'<<1, 'E'<<1, 'N'<<1, ' '<<1, ' '<<1, '0'<<1, /* OPEN   */
	'C'<<1, 'Q'<<1, ' '<<1, ' '<<1, ' '<<1, ' '<<1, '0'<<1, /* CQ     */
	'B'<<1, 'E'<<1, 'A'<<1, 'C'<<1, 'O'<<1, 'N'<<1, '0'<<1, /* BEACON */
	'R'<<1, 'M'<<1, 'N'<<1, 'C'<<1, ' '<<1, ' '<<1, '0'<<1, /* RMNC   */
	'A'<<1, 'L'<<1, 'L'<<1, ' '<<1, ' '<<1, ' '<<1, '0'<<1, /* ALL    */
	'F'<<1, 'L'<<1, 'X'<<1, 'N'<<1, 'E'<<1, 'T'<<1, '0'<<1, /* FLXNET */
	'L'<<1, 'I'<<1, 'N'<<1, 'K'<<1, 'S'<<1, ' '<<1, '0'<<1, /* LINKS  */
	'S'<<1, 'T'<<1, 'A'<<1, 'T'<<1, 'U'<<1, 'S'<<1, '0'<<1, /* STATUS */
	'R'<<1, 'O'<<1, 'U'<<1, 'T'<<1, 'E'<<1, ' '<<1, '0'<<1, /* ROUTE  */
	'\0',
};
char Mycall[AXALEN] = {
	'N'<<1, 'O'<<1, 'C'<<1, 'A'<<1, 'L'<<1, 'L'<<1, '0'<<1  /* NOCALL */
};
char Myalias[AXALEN] = {
	'T'<<1, 'C'<<1, 'P'<<1, 'I'<<1, 'P'<<1, ' '<<1, '0'<<1  /* TCPIP  */
};
char rttcheck[128] = "WAMPlus layer 2 link quality check";

struct ax_route *Ax_routes[AXROUTESIZE];
struct iface *Axroute_default_ifp;
struct axifpar basepar = {
		Mycall,       /* hwaddr 	*/
		Myalias,      /* alias  	*/
		NULLCHAR,     /* bcdest 	*/
		MODE_VC,      /* encap mode	*/
		0,            /* dama           */
		1,            /* ignore MOD128  */
		0,            /* ignore DAMA    */
		7,            /* maxframe       */
		256,          /* paclen		*/
		128,          /* pthresh        */
		20,           /* retries	*/
		5000,         /* t1		*/
		400,          /* t2		*/
		600000L,      /* t3		*/
		0,            /* t3disc		*/ /* superbloede angewohnheit*/
		3600000L,     /* t4		*/
		TT_AUTO,      /* timertype      */
		0,	      /* permui		*/
		V2,           /* version        */
		8192,         /* window         */
		2,            /* digipeat       */
		16            /* blimit		*/
	};
       
int bcinterval = 0;
int default_filtermode = 0;
int default_modulo = 0;
char bctext[256];
struct timer bctimer;
int axroutecnt;

int getfiltermode(
char *origin)
{
	struct ax_route *rt;
	
	if ((rt=ax_routeptr(origin,0))==0) return (default_filtermode);
	else {
	   if (rt->filtermode != FILTER_DEFAULT) return(rt->filtermode);
	   else return(default_filtermode);
	}
}

int getmodulo(
char *origin)
{
	struct ax_route *rt;
	
	if ((rt=ax_routeptr(origin,0))==0) return (default_modulo);
	else return(rt->modulo128);
}

void sendbc(
void)
{
	struct iface *ifp;
	struct mbuf  *sendbuf;
	char src[AXALEN];
	char dst[AXALEN];
	
	if (Ifaces!=NULLIF) {
	for (ifp=Ifaces;(ifp!=NULLIF);ifp=ifp->next) if (ifp->ax25.bcdest!=NULLCHAR) {
		sendbuf=alloc_mbuf(strlen(bctext));
		memcpy(sendbuf->data, bctext, strlen(bctext));
		sendbuf->cnt=strlen(bctext);		
		memcpy(src,ifp->ax25.hwaddr,AXALEN);
		memcpy(dst,ifp->ax25.bcdest,AXALEN);
		(*ifp->output)(ifp,dst,src,PID_NO_L3,sendbuf);
	}		 
	if (bcinterval!=0) {
	   set_timer(&bctimer, bcinterval*1000);
	   start_timer(&bctimer);
	} else stop_timer(&bctimer);
	}
}

void
restimerexp(
void *p)
{
	struct ax_route *axp = (struct ax_route *) p;
	stop_timer(axp->chktimer);
	free(axp->chktimer);
	if (axp->chkincnt<RESTRESHH) axp->rttchkres=RES_I;
	else axp->rttchkres=RES_UI;
	axp->chkincnt=0;
	axp->chkttl=RESPTTL;
}

void
sendresframes(
char *hwaddr,
struct iface *ifp)
{
	struct mbuf *bp;
	int i;
	for (i=0;i<=RESPCKTNUM;i++) {
	   bp=alloc_mbuf(132);
	   memcpy(bp->data,&rttcheck,128);
	   bp->cnt=128;
	   bp = pushdown(bp,1);
	   bp->data[0] = (char)0xF0;
	   axsend(ifp,hwaddr,ifp->ax25.hwaddr,LAPB_COMMAND,UI+PF,bp);
	}
	return;
}

int 
resolve_mode(
char *hw_addr, 
struct iface *ifp)
{
	struct ax_route *axp;
	struct ax25_cb *axcb;
	char temp[AXALEN];
	
	if (hw_addr==NULLCHAR) return (MODE_DATA);
	axp = ax_routeptr(hw_addr,0);
	if (axp==NULLAXR) return (MODE_DATA);
	if (find_ax25f(Mycall, hw_addr)) return (MODE_VC);
	if (axp->chkttl) {
	   axp->chkttl--;
	   if (axp->rttchkres==RES_UI) return (MODE_DATA);
	   if (axp->rttchkres==RES_I) return (MODE_VC);
	}
	if (axp->rttchkres==RES_INPROG) return (MODE_DATA);
	
	/* Green light for check */
	axp->rttchkres = RES_INPROG;
	axp->chkincnt = 0;
	axp->chkttl = 0;
	sendresframes(hw_addr,ifp);
	axp->chktimer=malloc(sizeof(struct timer));
	axp->chktimer->func = restimerexp;
	axp->chktimer->arg = axp;
	set_timer(axp->chktimer, RESTIME);
	start_timer(axp->chktimer);
	return(MODE_DATA);
}	

int
isuseraddr(
char *ucall)
{
	struct ax25_cb *axp = NULLAX25;
	if (Ax25_cb!=NULLAX25) {
	   if (addreq(Ax25_cb->hdr.source,ucall)) return(1);
	   for(axp = Ax25_cb; axp->next != NULLAX25; axp=axp->next) 
		if (addreq(axp->hdr.source,ucall)) return(1);
	} else return (0);   
        if (addreq(axp->hdr.source,ucall)) return(1);
        return(0);
}

/* Send IP datagrams across an AX.25 link */
int
ax_send(
struct mbuf *bp,
struct iface *iface,
int32 gateway,
int tos)
{
	struct mbuf *tbp;
	char *hw_addr;
	struct ax25_cb *axp;
	int prefmode;
	int prefmodeh;
	struct arp_tab *arpp; 

	if((hw_addr = res_arp(iface,ARP_AX25,gateway,bp)) == NULLCHAR)
		return 0;       /* Wait for address resolution */
	arpp=arp_lookup(ARP_AX25,gateway);
	if ((arpp->mode)==MODE_DEFAULT) prefmodeh=iface->ax25.ifmode;
	else prefmodeh=arpp->mode;
	if (prefmodeh==MODE_AUTO) prefmode=resolve_mode(hw_addr,iface);
	else prefmode=prefmodeh;
	/* UI frames are used for any one of the following three conditions:
	 * 1. The "low delay" bit is set in the type-of-service field.
	 * 2. The "reliability" TOS bit is NOT set and the interface is in
	 *    datagram mode.
	 * 3. The destination is the broadcast address (this is helpful
	 *    when broadcasting on an interface that's in connected mode).
	 */
	/* DG1KJD 5-4-95:
	   ...
	   only used if "ax25 useui" on
	   DG1KJD 10-4-95:
	   changed the whole story... starting with
	   encap names "KISSI" "KISSUI" and so on...
	 */
	 if(((tos & IP_COS) == DELAY
	 || ((tos & IP_COS) != RELIABILITY && (prefmode == MODE_DATA))
	 || addreq(hw_addr,Ax25multi[0]))){
		/* Use UI frame */
		if (!((prefmode == MODE_VC) && (!(iface->ax25.useui))))
		return (*iface->output)(iface,hw_addr,iface->ax25.hwaddr,PID_IP,bp);
	}
 	/* Reliability is needed; use I-frames in AX.25 connection */
	if((axp = find_ax25f(iface->ax25.hwaddr,hw_addr)) == NULLAX25){
		/* Open a new connection */
		struct ax25 hdr;
		hdr.ndigis = hdr.nextdigi = 0;
		addrcp(hdr.dest,hw_addr);
		addrcp(hdr.source,iface->ax25.hwaddr);
		axp = open_ax25(&hdr,
		 AX_ACTIVE,NULLVFP,NULLVFP,NULLVFP,NULLCHAR);
		if(axp == NULLAX25){
			free_p(bp);
			return -1;
		}
	}
	if(axp->state == LAPB_DISCONNECTED){
		est_link(axp);
		lapbstate(axp,LAPB_SETUP);
	}
	/* Insert the PID */
	bp = pushdown(bp,1);
	bp->data[0] = PID_IP;
	if((tbp = segmenter(bp,axp->paclen)) == NULLBUF){
		free_p(bp);
		return -1;
	}
	return send_ax25(axp,tbp,-1);
}
/* Add header and send connectionless (UI) AX.25 packet.
 * Note that the calling order here must match enet_output
 * since ARP also uses it.
 */
int
ax_output(
struct iface *iface,    /* Interface to use; overrides routing table */
char *dest,             /* Destination AX.25 address (7 bytes, shifted) */
char *source,           /* Source AX.25 address (7 bytes, shifted) */
uint16 pid,             /* Protocol ID */
struct mbuf *bp)        /* Data field (follows PID) */
{
	/* Prepend pid to data */
	bp = pushdown(bp,1);
	bp->data[0] = (char)pid;
	return axsend(iface,dest,source,LAPB_COMMAND,UI,bp);
}
/* Common subroutine for sendframe() and ax_output() */
static int
axsend(
struct iface *iface,    /* Interface to use; overrides routing table */
char *dest,             /* Destination AX.25 address (7 bytes, shifted) */
char *source,           /* Source AX.25 address (7 bytes, shifted) */
int cmdrsp,             /* Command/response indication */
int ctl,                /* Control field */
struct mbuf *bp)        /* Data field (includes PID) */
{
	struct ax25 addr;
	struct iface *ifp;
	char *idest;
	int rval;
	struct mbuf *data;

	/* If the source addr is unspecified, use the interface address */
	if(source[0] == '\0')
		source = iface->ax25.hwaddr;

	/* Do AX.25 routing */
	memcpy(addr.dest,dest,AXALEN);
	memcpy(addr.source,source,AXALEN);
	addr.ndigis = 0;
	axroute(&addr, &ifp);
	addr.cmdrsp = cmdrsp;

	if(addr.ndigis != 0 && addr.nextdigi != addr.ndigis){
		idest = addr.digis[addr.nextdigi];
	} else {
		idest = dest;
	}

	/* Allocate mbuf for control field, and fill in */
	bp = pushdown(bp,1);
	bp->data[0] = ctl;

	if((data = htonax25(&addr,bp)) == NULLBUF){
		free_p(bp);
		return -1;
	}
	/* This shouldn't be necessary because redirection has already been
	 * done at the IP router layer, but just to be safe...
	 */
	if(iface->forw != NULLIF){
		logsrc(iface->forw,iface->forw->ax25.hwaddr);
		logdest(iface->forw,idest);
		rval = (*iface->forw->raw)(iface->forw,data);
	} else {
		logsrc(iface,iface->ax25.hwaddr);
		logdest(iface,idest);
		rval = (*iface->raw)(iface,data);
	}
	return rval;
}
/* Process incoming AX.25 packets.
 * After optional tracing, the address field is examined. If it is
 * directed to us as a digipeater, repeat it.  If it is addressed to
 * us or to QST-0, kick it upstairs depending on the protocol ID.
 */
void
ax_recv(
struct iface *iface,
struct mbuf *bp)
{
	struct mbuf *hbp;
	char control;
	struct ax25 hdr;
	char (*mpp)[AXALEN];
	int mcast;
	char *isrc,*idest;      /* "immediate" source and destination */
	struct ax25_cb *axp;

	/* Pull header off packet and convert to host structure */
	if(ntohax25(&hdr,&bp,iface) < 0){
		/* Something wrong with the header */
		iface->ax25errors++;
		free_p(bp);
		return;
	}
	/* If there were digis in this packet and at least one has
	 * been passed, then the last passed digi is the immediate source.
	 * Otherwise it is the original source.
	 */
	if(hdr.ndigis != 0 && hdr.nextdigi != 0)
		isrc = hdr.digis[hdr.nextdigi-1];
	else
		isrc = hdr.source;

	/* If there are digis in this packet and not all have been passed,
	 * then the immediate destination is the next digi. Otherwise it
	 * is the final destination.
	 */
	if(hdr.ndigis != 0 && hdr.nextdigi != hdr.ndigis)
		idest = hdr.digis[hdr.nextdigi];
	else
		idest = hdr.dest;

	/* Don't log our own packets if we overhear them, as they're
	 * already logged by axsend() and by the digipeater code.
	 */
	if( !(addreq(isrc,iface->ax25.hwaddr) || addreq(isrc,iface->ax25.alias)) ){
		logsrc(iface,isrc);
		logdest(iface,idest);
	}
	/* Examine immediate destination for a multicast address */
	mcast = 0;
	for(mpp = Ax25multi;(*mpp)[0] != '\0';mpp++){
		if(addreq(idest,*mpp)){
			mcast = 1;
			break;
		}
	}
	if(!mcast && !addreq(idest,iface->ax25.hwaddr) && !addreq(idest, iface->ax25.alias)
	&& !isuseraddr(idest)){
		/* Not a broadcast, and not addressed to us. Inhibit
		 * transmitter to avoid colliding with addressed station's
		 * response, and discard packet.
		 */
#if 0
		if(iface->ioctl != NULL)
			(*iface->ioctl)(iface,PARAM_MUTE,1,-1);
#endif
		free_p(bp);
		return;
	}
#if 0
	if(!mcast && iface->ioctl != NULL){
		/* Packet was sent to us; abort transmit inhibit */
		(*iface->ioctl)(iface,PARAM_MUTE,1,0);
	}
#endif
	/* At this point, packet is either addressed to us, or is
	 * a multicast.
	 */
	axroute_add(iface, &hdr, 0);
	if(hdr.nextdigi < hdr.ndigis){
		/* Packet requests digipeating. See if we can repeat it. */
		if(iface->ax25.Digipeat && !mcast){
			/* Yes, kick it back out. htonax25 will set the
			 * repeated bit.
			 */
			hdr.nextdigi++;
			if(iface->ax25.Digipeat == 1 ||
			   (bp && (*bp->data & ~PF) == UI) ||
			   addreq(hdr.source, hdr.dest)) {
				struct iface *ifp;
				axroute(&hdr, &ifp);
				if((hbp = htonax25(&hdr,bp)) != NULLBUF){
					bp = hbp;
					if (ifp) {
						logsrc(ifp,ifp->ax25.hwaddr);
						logdest(ifp,hdr.nextdigi != hdr.ndigis ? hdr.digis[hdr.nextdigi] : hdr.dest);
						(*ifp->raw)(ifp, bp);
						bp = NULLBUF;
					}
				}
			} else {
				lapb_input(iface,&hdr,bp);
				bp = NULLBUF;
			}
		}
		free_p(bp);     /* Dispose if not forwarded */
		return;
	}
	/* If we reach this point, then the packet has passed all digis,
	 * and is either addressed to us or is a multicast.
	 */
	if(bp == NULLBUF)
		return;         /* Nothing left */

	/* Sneak a peek at the control field. This kludge is necessary because
	 * AX.25 lacks a proper protocol ID field between the address and LAPB
	 * sublayers; a control value of UI indicates that LAPB is to be
	 * bypassed.
	 */
	setcall(hdr.dest,iface->ax25.hwaddr);   	/* Frame might still be adressed    */
	setcall(idest,iface->ax25.hwaddr);		/* to aliasaddress. Correct this.   */
						/* LAPB code will take care of      */
						/* using the right source address   */
						/* by extracting the header on its  */
						/* own. Also needed by "connect     */
						/* patch. Quick, but effective hack.*/	
	control = *bp->data;
	if(uchar(control & ~PF) == UI){
		int pid;
		struct axlink *ipp;

		if ((uchar(control) & PF) && ((axp=find_ax25f(iface->ax25.hwaddr,hdr.source))==0)) axsend(iface,hdr.source,iface->ax25.hwaddr,LAPB_RESPONSE,DM,NULLBUF);
		(void) PULLCHAR(&bp);
		if((pid = PULLCHAR(&bp)) == -1)
			return;         /* No PID */
		/* Find network level protocol and hand it off */
		for(ipp = Axlink;ipp->funct != NULL;ipp++){
			if(ipp->pid == pid)
				break;
		}
		if(ipp->funct != NULL)
			(*ipp->funct)(iface,NULLAX25,hdr.source,hdr.dest,bp,mcast);
		else
			free_p(bp);
		return;
	}
	/* Everything from here down is connected-mode LAPB, so ignore
	 * multicasts
	 */
	if(mcast){
		free_p(bp);
		return;
	}

	lapb_input(iface,&hdr,bp);
}

/* General purpose AX.25 frame output */
int
sendframe(
struct ax25_cb *axp,
int cmdrsp,
int ctl,
struct mbuf *data)
{
	struct mbuf *bp;
	struct iface *ifp;

	bp = pushdown(data,1);
	bp->data[0] = ctl;
	axp->hdr.cmdrsp = cmdrsp;
	if((data = htonax25(&axp->hdr,bp)) == NULLBUF){
		free_p(bp);
		return -1;
	}
	if ((ifp = axp->iface)) {
		if (ifp->forw)
			ifp = ifp->forw;
		logsrc(ifp,ifp->ax25.hwaddr);
		logdest(ifp,axp->hdr.nextdigi != axp->hdr.ndigis ? axp->hdr.digis[axp->hdr.nextdigi] : axp->hdr.dest);
		return (*ifp->raw)(ifp,data);
	}
	free_p(data);
	return -1;
}

/* General purpose AX.25 frame output */
int
sendframeE(
struct ax25_cb *axp,
int cmdrsp,
int ctl1,
int ctl2,
struct mbuf *data)
{
	struct mbuf *bp;
	struct iface *ifp;

	bp = pushdown(data,2);
	bp->data[0] = ctl1;
	bp->data[1] = ctl2;
	axp->hdr.cmdrsp = cmdrsp;
	if((data = htonax25(&axp->hdr,bp)) == NULLBUF){
		free_p(bp);
		return -1;
	}
	if ((ifp = axp->iface)) {
		if (ifp->forw)
			ifp = ifp->forw;
		logsrc(ifp,ifp->ax25.hwaddr);
		logdest(ifp,axp->hdr.nextdigi != axp->hdr.ndigis ? axp->hdr.digis[axp->hdr.nextdigi] : axp->hdr.dest);
		return (*ifp->raw)(ifp,data);
	}
	free_p(data);
	return -1;
}

int
valid_remote_call(
const char *call)
{
	char (*mpp)[AXALEN];

	if (!*call || ismyax25addr(call))
		return 0;
	for (mpp = Ax25multi; (*mpp)[0]; mpp++)
		if (addreq(*mpp, call))
			return 0;
	return 1;
}

static int
axroute_hash(
const char *call)
{
	int hashval;

	hashval  = ((*call++ << 23) & 0x0f000000);
	hashval |= ((*call++ << 19) & 0x00f00000);
	hashval |= ((*call++ << 15) & 0x000f0000);
	hashval |= ((*call++ << 11) & 0x0000f000);
	hashval |= ((*call++ <<  7) & 0x00000f00);
	hashval |= ((*call++ <<  3) & 0x000000f0);
	hashval |= ((*call   >>  1) & 0x0000000f);
	return hashval;
}

struct ax_route *
ax_routeptr(
const char *call,
int create)
{

	struct ax_route **tp;
	struct ax_route *rp;

	tp = Ax_routes + (axroute_hash(call) % AXROUTESIZE);
	for (rp = *tp; rp && !addreq(rp->target, call); rp = rp->next)
		;
	if (!rp && create) {
		axroutecnt++;
		rp = (struct ax_route *) calloc(1, sizeof(struct ax_route));
		addrcp(rp->target, call);
		rp->next = *tp;
		*tp = rp;
	}
	return rp;
}

void
axroute_add(
struct iface *iface,
struct ax25 *hdr,
int perm)
{

	char *call;
	char *calls[MAXDIGIS+1];
	int i;
	int j;
	int ncalls = 0;
	struct ax_route *lastnode = 0;
	struct ax_route *rp;
	int savemod;
	int savemode;
	int saverttchkres;
	int savechkincnt;
	int savechkttl;
	struct timer *savetimer;

	call = hdr->source;
	if (!valid_remote_call(call))
		return;
	calls[ncalls++] = call;
	for (i = 0; i < hdr->nextdigi; i++) {
		call = hdr->digis[i];
		if (!valid_remote_call(call))
			return;
		for (j = 0; j < ncalls; j++)
			if (addreq(call, calls[j]))
				return;
		calls[ncalls++] = call;
	}

	for (i = ncalls - 1; i >= 0; i--) {
		savemode=FILTER_DEFAULT;
		savemod=default_modulo;
		saverttchkres=RES_I;
		savechkincnt=0;
		savechkttl=0;
		savetimer=NULLTIMER;
		if ((rp = ax_routeptr(calls[i], 0))!=0) {
		   savemode=(rp->filtermode);
		   savemod=(rp->modulo128);
		   saverttchkres=(rp->rttchkres);
		   savechkincnt=(rp->chkincnt);
		   savechkttl=(rp->chkttl);
		   savetimer=NULLTIMER;
		}
		rp = ax_routeptr(calls[i], 1);
		if (perm || !rp->perm) {
			if (lastnode) {
				rp->digi = lastnode;
				rp->ifp = 0;
			} else {
				rp->digi = 0;
				rp->ifp = iface;
			}
			rp->perm = perm;
		}
		rp->time = secclock();
		rp->filtermode = savemode;
		rp->modulo128 = savemod;
		rp->rttchkres = saverttchkres;
		rp->chkincnt = savechkincnt;
		rp->chkttl = savechkttl;
		rp->chktimer = savetimer;
		if (savetimer) savetimer->arg=rp;
		lastnode = rp;
	}
	axroute_savefile();
}

void
axroute(
struct ax25 *hdr,
struct iface **ifpp)
{

	char *idest;
	int d;
	int i;
	struct ax_route *rp;
	struct iface *ifp;

	/*** Find my last address ***/

	hdr->nextdigi = 0;
	for (i = hdr->ndigis - 1; i >= 0; i--)
		if (ismyax25addr(hdr->digis[i])) {
			hdr->nextdigi = i + 1;
			break;
		}

	/*** Remove all digipeaters before me ***/

	d = hdr->nextdigi - 1;
	if (d > 0) {
		for (i = d; i < hdr->ndigis; i++)
			addrcp(hdr->digis[i-d], hdr->digis[i]);
		hdr->ndigis -= d;
		hdr->nextdigi = 1;
	}

	/*** Add necessary digipeaters and find interface ***/

	ifp = 0;
	idest = hdr->nextdigi < hdr->ndigis ? hdr->digis[hdr->nextdigi] : hdr->dest;
	for (rp = ax_routeptr(idest, 0); rp; rp = rp->digi) {
		if (rp->digi && hdr->ndigis < MAXDIGIS) {
			for (i = hdr->ndigis - 1; i >= hdr->nextdigi; i--)
				addrcp(hdr->digis[i+1], hdr->digis[i]);
			hdr->ndigis++;
			addrcp(hdr->digis[hdr->nextdigi], rp->digi->target);
		}
		ifp = rp->ifp;
	}
	if (!ifp)
		ifp = Axroute_default_ifp;
	if (ifp && ifp->forw)
		ifp = ifp->forw;
	*ifpp = ifp;

	/*** Replace my address with hwaddr of interface ***/

	if (ifp)
		addrcp(hdr->nextdigi ? hdr->digis[0] : hdr->source, hdr->nextdigi ? ifp->ax25.hwaddr : hdr->source);
}

/* Handle ordinary incoming data (no network protocol) */
void
axnl3(
struct iface *iface,
struct ax25_cb *axp,
char *src,
char *dest,
struct mbuf *bp,
int mcast)
{
	if(axp == NULLAX25){
		free_p(bp);
	} else {
	   if (!(getfiltermode(src) & (FILTER_AX25+DISC_AX25))) {
	   	append(&axp->rxq,bp);
		if(axp->r_upcall != NULLVFP)
			(*axp->r_upcall)(axp,len_p(axp->rxq));
	   } else {
	      free_p(bp);  /* into trashcan */
	      if (getfiltermode(src) & DISC_AX25) disc_ax25(axp); /* ...see ya */
	   }
	}
}
