/* @(#) $Header: axserver.c,v 1.12 94/10/06 16:15:21 deyke Exp $ */

#include "global.h"
#include "mbuf.h"
#include "ax25.h"
#include "lapb.h"
#include "login.h"

int Axserver_enabled;
static void axserv_recontimeout(void *p);

/*---------------------------------------------------------------------------*/

static void axserv_recv_upcall(struct ax25_cb *axp, int cnt)
{
  struct mbuf *bp;

  bp = recv_ax25(axp, 0);
  if (bp) login_write((struct login_cb *) axp->user, bp);
}

/*---------------------------------------------------------------------------*/

static void axserv_send_upcall(struct ax25_cb *axp, int cnt)
{
  struct mbuf *bp;

  if ((bp = login_read((struct login_cb *) axp->user, space_ax25(axp))))
    send_ax25(axp, bp, PID_NO_L3);
}

/*---------------------------------------------------------------------------*/

static void axserv_state_upcall(struct ax25_cb *axp, int oldstate, int newstate)
{
  struct mbuf *bp;
  struct login_cb *tp = (struct login_cb *) axp->user;

  if (login_recon) {
    if (newstate == LAPB_DISCONNECTED && (!(tp->wantclose))) {
      if ((oldstate == LAPB_CONNECTED) || (oldstate == LAPB_RECOVERY)) {
         set_timer(tp->rectimer, Rectimeout);
         start_timer(tp->rectimer);
      }
      est_link(axp);
      lapbstate(axp, LAPB_SETUP);
      tp->recon_inprog = 1;
      return;
    }
    if ((newstate == LAPB_DISCONNECTED) && (tp->wantclose)) {
      login_close((struct login_cb *) axp->user);
      del_ax25(axp);
      return;
    }
    if ((newstate == LAPB_CONNECTED) && (tp->recon_inprog)) {
      bp = alloc_mbuf(1024);
      sprintf(bp->data,"[%s] - AutoReconnect.\r", Version);
      bp->cnt = strlen(bp->data);
      send_ax25(axp, bp, PID_NO_L3);
      stop_timer(tp->rectimer);
      tp->recon_inprog = 0;
      return;
    }
  } else {
    if (newstate == LAPB_DISCONNECTED) {
      login_close((struct login_cb *) axp->user);
      del_ax25(axp);
      return;
    }
  }
}

/*---------------------------------------------------------------------------*/

static void axserv_recontimeout(
void *p)
{
  struct ax25_cb *axp = (struct ax25_cb *) p;
  struct login_cb *tp = (struct login_cb *) axp->user;
  
  login_close(tp);
  disc_ax25(axp);
  del_ax25(axp);
}

/*---------------------------------------------------------------------------*/

void axserv_open(struct ax25_cb *axp, int cnt)
{
  char callsign[AXBUF];

  if (Axserver_enabled) {
    pax25(callsign, axp->hdr.dest);
    axp->user = (char *) login_open(callsign, "AX25", (void (*)(void *)) axserv_send_upcall, (void (*)(void *)) disc_ax25, axp);
  }
  if (axp->user) {
    free_q(&axp->rxq);
    axp->r_upcall = axserv_recv_upcall;
    axp->t_upcall = axserv_send_upcall;
    axp->s_upcall = axserv_state_upcall;
    ((struct login_cb*)axp->user)->rectimer->func = axserv_recontimeout;
    ((struct login_cb*)axp->user)->rectimer->arg = axp;
    set_timer(((struct login_cb*)axp->user)->rectimer, Rectimeout);
    stop_timer(((struct login_cb*)axp->user)->rectimer);
  } else
    disc_ax25(axp);
}

/*---------------------------------------------------------------------------*/

int ax25start(int argc, char *argv[], void *p)
{
  Axserver_enabled = 1;
  return 0;
}

/*---------------------------------------------------------------------------*/

int ax250(int argc, char *argv[], void *p)
{
  Axserver_enabled = 0;
  return 0;
}
