/* connect.c  -  WAMPlus support program. Gives users the ability to connect
   out with their own callsigns - DG1KJD 21-9-95 . Basic code from dk5sg's
   cnet.c
*/
   
#include <sys/types.h>
#include <curses.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <unistd.h>
#include <pwd.h>
#include <termios.h>
#define SEL_ARG(x) (x)
#include "../lib/buildsaddr.h"
#include "../lib/callvalid.h"

struct mbuf {
  struct mbuf *next;
  unsigned int cnt;
  char *data;
};

char *dest;
int conn = 0;
int inecho = 0;
static int Ansiterminal = 1;
static int fdin = 0;
static int fdout = 1;
static int fdsock = -1;
static struct mbuf *sock_queue;
static struct mbuf *term_queue;

static struct termios prev_termios;

static void terminate(void);
static void recvq(int fd, struct mbuf **qp);
static void sendq(int fd, struct mbuf **qp);

static void terminate(void)
{
  close(fdsock);
  fcntl(fdout, F_SETFL, fcntl(fdout, F_GETFL, 0) & ~O_NONBLOCK);
  for (; term_queue; term_queue = term_queue->next)
    write(fdout, term_queue->data, term_queue->cnt);
  tcsetattr(fdin, TCSANOW, &prev_termios);
  exit(0);
}

static void recvq(int fd, struct mbuf **qp)
{

  char buf[1024];
  int n;
  struct mbuf *bp, *tp;

  n = read(fd, buf, sizeof(buf));
  if ((!conn) && (fd==fdin) && ((buf[0]=='Q') || (buf[0]=='q'))) {
    printf("*** failure with %s\n",dest);
    terminate();
  }
  if (n <= 0) {
     if (conn) printf("\n*** DISC received from %s\n*** reconnected\n",dest);
     terminate();
  }
  bp = (struct mbuf *) malloc(sizeof(struct mbuf) + n);
  if (!bp) {
     if (conn) printf("\n*** Out of memory. Disconnecting from %s\n",dest);
     terminate();
  }
  bp->next = 0;
  bp->cnt = n;
  bp->data = (char *) (bp + 1);
  memcpy((bp->data), buf, n);
  if ((!conn) && (fd==fdsock)) {
    printf("*** connected to %s\n",dest);
    conn=1;
  }
  if (*qp) {
    for (tp = *qp; tp->next; tp = tp->next) ;
    tp->next = bp;
  } else
    *qp = bp;
}

static void sendq(int fd, struct mbuf **qp)
{
  int n;
  struct mbuf *bp;

  bp = *qp;
/*  if ((fd!=fdout) && inecho) write(fdout, bp->data, bp->cnt); */
  n = write(fd, bp->data, bp->cnt);
  if (n <= 0) {
     if (conn) printf("\n*** DISC received from %s\n*** reconnected\n",dest);
     terminate();
  }
  bp->data += n;
  bp->cnt -= n;
  if (!bp->cnt) {
    *qp = bp->next;
    free(bp);
  }
}

void strupcase(char *str)
{
   int i;
   i=0;
   while (str[i]!=0) {
      str[i]=toupper(str[i]);
      i++;
   }
}

void scanargs(int *argc, char **argv)
{
   if (argc>0) if ((strcmp(argv[(*argc)-1],"-e")==0) ||
                   (strcmp(argv[(*argc)-1],"-E")==0)) {
      inecho=1;
      (*argc)--;
   }
} 

int main(int argc, char **argv)
{
  char sstr[128];
  char *ap;
  char *server;
  char area[1024];
  char bp[1024];
  int addrlen;
  int flags;
  struct fd_set rmask;
  struct fd_set wmask;
  struct sockaddr *addr;
  int uid;
  struct passwd *pwd;
  struct termios curr_termios;

  printf("Connect V1.3 - by DG1KJD 21-9-95\n");
  scanargs(&argc, argv);
  if (argc>1) {

  uid=getuid();
  pwd=getpwuid(uid);
  strupcase(argv[1]);
  tcgetattr(fdin, &prev_termios);
  ap = area;
  Ansiterminal = 1;

  server = "unix:/tcp/sockets/netusr";
  if (!(addr = build_sockaddr(server, &addrlen))) {
    fprintf(stderr, "Could not build socket address!\n");
    terminate();
  }
  if ((fdsock = socket(addr->sa_family, SOCK_STREAM, 0)) < 0) {
    perror("socket");
    terminate();
  }
  if (connect(fdsock, addr, addrlen)) {
    perror("connect");
    terminate();
  }
  if (!callvalid(pwd->pw_name)) {
    printf("\nYour login name is not a valid amateur radio callsign. Go away.\n");
    terminate();
  }
  printf("\nlink setup... type Q to abort.\n");
  sprintf(sstr, "connect ax25x %s-15 %s",pwd->pw_name,argv[1]);
  dest=argv[1];
  argv++;
  argc--;
  while (--argc) {
     argv++;
     if (strncmp("via",*argv,strlen(*argv))) {
        strcat(sstr," ");
        strcat(sstr,*argv);
     }
  }
  strcat(sstr,"\n");
  write(fdsock,sstr, strlen(sstr));

  if ((flags = fcntl(fdsock, F_GETFL, 0)) == -1 ||
       fcntl(fdsock, F_SETFL, flags | O_NONBLOCK) == -1) {
    perror("fcntl");
    terminate();
  }

  curr_termios = prev_termios;
  curr_termios.c_lflag = ICANON | (inecho ? ECHO : 0);
  curr_termios.c_cc[VMIN] = 0;
  curr_termios.c_cc[VTIME] = 0;
  tcsetattr(fdin, TCSANOW, &curr_termios);
  if ((flags = fcntl(fdout, F_GETFL, 0)) == -1 ||
      fcntl(fdout, F_SETFL, flags | O_NONBLOCK) == -1) {
      perror("fcntl");
      terminate();
  }

  for (; ; ) {
    FD_ZERO(&rmask);
    FD_SET(fdin, &rmask);
    FD_SET(fdsock, &rmask);
    FD_ZERO(&wmask);
    if (sock_queue) FD_SET(fdsock, &wmask);
    if (term_queue) FD_SET(fdout,  &wmask);
    if (select(fdsock + 1, SEL_ARG(&rmask), SEL_ARG(&wmask), SEL_ARG(0), (struct timeval *) 0) > 0) {
      if (FD_ISSET(fdsock, &rmask)) recvq(fdsock, &term_queue);
      if (FD_ISSET(fdin,   &rmask)) recvq(fdin,   &sock_queue);
      if (FD_ISSET(fdsock, &wmask)) sendq(fdsock, &sock_queue);
      if (FD_ISSET(fdout,  &wmask)) sendq(fdout,  &term_queue);
  }  
  }
 } else {
    printf("Usage: connect <dest> [via] [<digi1>] [<digi2>] ... [-E]\n");
    return 1;
 }
 return 0;
}
